// ====================================================================
// Copyright (C) 2011 - Tsubasa Saito
// Copyright (C) 2018 - Samuel GOUGEON
//
// This file must be used under the terms of the CeCILL.
// This source file is licensed as described in the file COPYING, which
// you should have received as part of this distribution.  The terms
// are also available at
// http://www.cecill.info/licences/Licence_CeCILL_V2-en.txt
// ====================================================================

// *****************************************************************
// Norms of a matrix of DD numbers
//
// Calling Sequence :
// norm(x,N)
//
// Arguments :
// x : DD matrix (or column, for N=2)
// N : 1 or 2 or %inf or 'fro'
//
// ***TODO : extend norm(*,2) to matrices (restricted to columns)
//
// Authors :
// Copyright (C) 2011 - Tsubasa Saito
// Copyright (C) 2018 - Samuel GOUGEON
// last updating : 2018-07-31
// *****************************************************************
function z = %dd_norm(x,N)
    [m,n] = size(x.hi);
    if ~isdef("N","l") then
        N = 1
    elseif type(N)==10
        N = part(convstr(strcat(stripblanks(N))),1)
    end
    if m==1 then
        x = x'
        [m, n] = (n, m)
    end

    // Processing
    // ----------
    select N
    case 1 // 1-norm
        x = abs(x);
        z = 0;
        for i = 1:m
            z = z + x(i,:);
        end
        z = %dd_max(z)
        
    case 2 // 2-norm
        if n ~= 1 then
            msg = _("%s: Argument #%d: Column expected.\n")
            error(msprintf(msg, "%dd_norm", 1));
        end
        t = 134217729 * x.hi;
        ah = t - (t - x.hi);
        al = x.hi - ah;

        p1 = x.hi .* x.hi;
        p2 = ((ah .* ah - p1) + 2.0 * (ah .* al)) + al .* al;
        p2 = p2 + 2.0 * (x.hi .* x.lo);

        chi = p1 + p2;
        clo = p2 - (chi - p1);

        for k=2:m 
            sh = chi(k-1) + chi(k);
            v = sh - chi(k-1);
            eh = (chi(k-1) - (sh - v)) + (chi(k) - v);

            sl = clo(k-1) + clo(k);
            eh = eh + sl;
            chi(k) = sh + eh;
            clo(k) = eh - (chi(k) - sh);
        end

        c = sqrt(chi(m)); // initial value   
        p1 = c*c;
        t = 134217729 * c;
        ah = t - (t - c);
        al = c - ah;
        p2 = ((ah * ah - p1) + 2.0 * ah * al) + al * al;
        cc = (chi(m) - p1 - p2 + clo(m)) / (c * 2.0); // Newton Iterative

        tmp =  c + cc;
        z = dd(tmp, cc-(tmp-c));

    case %inf // inf-norm
        z = abs(x);
        x_max = z(:,1)
        for i = 2:n
            x_max = x_xmax + z(:,i)
        end
        z = %dd_max(x_max)

    case 'f'  // Frobenius
        z = 0;
        for i = 1:m
            tmp =  x(i,:)
            z = z + tmp * tmp'
        end
        z = sqrt(z)
    else
        msg = _("%s: Argument #%d: Must be in the set {%s}.\n")
        error(msprintf(msg, "%dd_norm", 2, "1,2,%inf,''fro''"));
    end
endfunction

