// ====================================================================
// Copyright (C) 2011 - Tsubasa Saito
//
// This file must be used under the terms of the CeCILL.
// This source file is licensed as described in the file COPYING, which
// you should have received as part of this distribution.  The terms
// are also available at
// http://www.cecill.info/licences/Licence_CeCILL_V2-en.txt
// ====================================================================

// *****************************************************************
// Tangent function for DD
//
// Calling Sequence :
// tan(x)
//
// Arguments :
// x : DD number
// 
// This function is based on Bailey's QD, dd_real tan
// http://crd.lbl.gov/~dhbailey/mpdist/
//
// Authors :
// Copyright (C) 2011 - Tsubasa Saito
// make this file : 2011/09/02
// last updating : 2011/09/02
// *****************************************************************
function t = %dd_tan(x)
    [sin_x, cos_x] = sincos(x);
    t = sin_x / cos_x;
endfunction

function [sin_x,cos_x] = sincos(x)

    if (x == 0) then
        sin_x = 0.0;
        cos_x = 1.0;
        return;
    end

    _2pi = dd(6.283185307179586232D+00,2.449293598294706414e-16);  // 2*pi
    _pi2 = dd(1.570796326794896558e+00,6.123233995736766036e-17);    // pi/2
    _pi16 = dd(1.963495408493620697e-01,7.654042494670957545e-18); // pi/16

    sin_table = dd([1.950903220161282758e-01;  3.826834323650897818e-01;
                    5.555702330196021776e-01;  7.071067811865475727e-01],...
                   [-7.991079068461731263e-18; -1.005077269646158761e-17;
                     4.709410940561676821e-17; -4.833646656726456726e-17]);
    cos_table = dd([9.807852804032304306e-01; 9.238795325112867385e-01;
                    8.314696123025452357e-01; 7.071067811865475727e-01],...
                   [1.854693999782500573e-17; 1.764504708433667706e-17;
                    1.407385698472802389e-18;-4.833646656726456726e-17]);

    // approximately reduce modulo 2*pi
    z = nint_dd(x / _2pi);
    r = x - (_2pi * z);

    // approximately reduce module pi/2 and pi/16
    j = floor(r.hi / _pi2.hi + 0.5);
    t = r - (_pi2 * j);
    abs_j = abs(j);
    k = floor(t.hi / _pi16.hi + 0.5);
    t = t - (_pi16 * k);
    abs_k = abs(k);

    if (abs_j > 2) then
        error("(sincos): Cannot reduce modulo pi/2.");
        // sin_x = tlist(['dd','hi','lo'],%nan,%nan);
        // cos_x = tlist(['dd','hi','lo'],%nan,%nan);
    end

    if (abs_k > 4) then
        error("(sincos): Cannot reduce modulo pi/16.");
        // sin_x = tlist(['dd','hi','lo'],%nan,%nan);
        // cos_x = tlist(['dd','hi','lo'],%nan,%nan);
    end

    [sin_t, cos_t] = sincos_taylor_dd(t);

    if (abs_k == 0) then
        s = sin_t;
        c = cos_t;
    else
        u = cos_table(abs_k,1);
        v = sin_table(abs_k,1);
        if (k > 0) then
            s = u * sin_t + v * cos_t;
            c = u * cos_t - v * sin_t;
        else
            s = u * sin_t - v * cos_t;
            c = u * cos_t + v * sin_t;
        end
    end

    if (abs_j == 0) then
        sin_x = s;
        cos_x = c;
    elseif (j == 1) then
        sin_x = c;
        cos_x = -s;
    elseif (j == -1) then
        sin_x = -c;
        cos_x = s;
    else
        sin_x = -s;
        cos_x = -c;
    end
endfunction
